package maze.gui;

import java.util.EnumMap;
import java.util.HashMap;

import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.Label;
import javafx.scene.control.RadioButton;
import javafx.scene.control.TextField;
import javafx.scene.control.ToggleGroup;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.shape.Polygon;
import javafx.scene.shape.Rectangle;
import maze.model.Cell;
import maze.model.Move;
import maze.model.Position;
import maze.model.Puzzle;
import maze.model.Trail;
import maze.searchers.Queue;
import maze.searchers.Searcher;
import maze.searchers.Stack;
import maze.searchers.WrappedQueue;
import maze.searchers.WrappedStack;

public class MazeController {
	@FXML
	TextField width;
	@FXML
	TextField height;
	@FXML
	Pane maze;
	@FXML
	CheckBox placeExplorer;
	@FXML
	CheckBox placeGoal;
	@FXML
	CheckBox clearFill;
	
	@FXML
	Button forward;
	@FXML
	Button left;
	@FXML
	Button right;
	
	Puzzle mazeData;
	Polygon ex;
	HashMap<Position,Rectangle> cells = new HashMap<>();
	double cellWidth, cellHeight;
	
	@FXML
	ChoiceBox<String> stackBox;
	@FXML
	ChoiceBox<String> queueBox;
	@FXML
	RadioButton stackButton;
	@FXML
	RadioButton queueButton;
	ToggleGroup stackQueue = new ToggleGroup();
	ClassFinder<Queue<Trail>> queueFinder = new ClassFinder<>(Queue.class, "maze.model.searchers");
	ClassFinder<Stack<Trail>> stackFinder = new ClassFinder<>(Stack.class, "maze.model.searchers");
	
	@FXML
	HBox open;
	@FXML
	HBox closed;
	@FXML
	HBox visited;
	EnumMap<Cell,HBox> cellBoxes = new EnumMap<>(Cell.class);
	EnumMap<Cell,TextField> cellCounts = new EnumMap<>(Cell.class);

	@FXML
	TextField steps;
	
	@FXML
	public void initialize() {
		width.setText("15");
		height.setText("10");
		
		setupButton(forward, Move.FORWARD);
		setupButton(left, Move.LEFT);
		setupButton(right, Move.RIGHT);
		
		clearFill.setSelected(true);
		
		clearFill.setOnAction(event -> {placeGoal.setSelected(false);placeExplorer.setSelected(false);});
		placeGoal.setOnAction(event -> {placeExplorer.setSelected(false);clearFill.setSelected(false);});
		placeExplorer.setOnAction(event -> {placeGoal.setSelected(false);clearFill.setSelected(false);});
	
		setupChoices(stackButton, stackBox, stackFinder);
		setupChoices(queueButton, queueBox, queueFinder);
		stackQueue.selectToggle(stackButton);
		
		cellBoxes.put(Cell.OPEN, open);
		cellBoxes.put(Cell.CLOSED, closed);
		cellBoxes.put(Cell.VISITED, visited);
		for (Cell type: Cell.values()) {
			cellCounts.put(type, new TextField());
			cellCounts.get(type).setEditable(false);
			cellBoxes.get(type).getChildren().add(new Label(type.toString()));
			cellBoxes.get(type).getChildren().add(cellCounts.get(type));
		}
		
		steps.setEditable(false);
	}
	
	<T> void setupChoices(RadioButton button, ChoiceBox<String> choices, ClassFinder<T> classes) {
		stackQueue.getToggles().add(button);
		for (String className: classes.getTypeNames()) {
			choices.getItems().add(className);
		}
		if (choices.getItems().size() > 0) {
			choices.getSelectionModel().select(0);
		}
	}
	
	void setupButton(Button b, Move m) {
		b.setOnAction(event -> {
			if (mazeData.hasExplorer()) {
				mazeData.moveExplorer(m);
				placeFigure();
			}
		});
	}
	
	@FXML
	public void resetMaze() {
		int w = Integer.parseInt(width.getText());
		int h = Integer.parseInt(height.getText());
		mazeData = new Puzzle(w, h);
		maze.getChildren().clear();
		
		cellWidth = maze.getWidth() / w;
		cellHeight = maze.getHeight() / h;
		
		for (int x = 0; x < w; x++) {
			for (int y = 0; y < h; y++) {
				Rectangle cell = new Rectangle(x * cellWidth, y * cellHeight, cellWidth, cellHeight);
				Position spot = new Position(x, y);
				cells.put(spot, cell);
				cell.setFill(mazeData.getColorFor(spot));
				cell.setOnMouseClicked(event -> {
					if (placeExplorer.isSelected()) {
						if (mazeData.canEnter(spot)) {
							mazeData.placeExplorer(spot);
							maze.getChildren().remove(ex);
							ex = new Polygon(new double[]{cellWidth/4, cellHeight/4, 3*cellWidth/4, cellHeight/4, cellWidth/2, 3*cellHeight/4});
							ex.setFill(Color.GREEN);
							maze.getChildren().add(ex);
							placeFigure();
						}
					} else if (placeGoal.isSelected()) {
						mazeData.placeGoal(spot);
						cells.get(spot).setFill(Color.ORANGE);
					} else {
						if (mazeData.canEnter(spot)) {
							mazeData.fill(spot);
						} else {
							mazeData.clear(spot);
						}
						cell.setFill(mazeData.getColorFor(spot));
					}
				});
				maze.getChildren().add(cell);
			}
		}
		updateCellCounts();
	}
	
	void placeFigure() {
		Position spot = mazeData.getExplorerPosition();
		ex.setTranslateX(spot.getX() * cellWidth);
		ex.setTranslateY(spot.getY() * cellHeight);
		ex.setRotate(mazeData.getExplorerHeading().getRotation());
	}
	
	@FXML
	void randomize() {
		resetMaze();
		mazeData.tunnelRandomly();
		recolor();
	}
	
	void recolor() {
		for (int i = 0; i < mazeData.getWidth(); i++) {
			for (int j = 0; j < mazeData.getHeight(); j++) {
				Position p = new Position(i, j);
				cells.get(p).setFill(mazeData.getColorFor(p));
			}
		}
		updateCellCounts();
	}
	
	@FXML
	void solve() {
		if (mazeData == null) {
			steps.setText("No maze created");
			return;
		}
		mazeData.clearVisits();
		if (!mazeData.hasExplorer()) {
			steps.setText("No explorer placed");
		} else if (!mazeData.hasGoal()) {
			steps.setText("No goal placed");
		} else {
			try {
				Trail solution = mazeData.solve(makeSearcher());
				recolor();
				if (solution == null) {
					steps.setText("Failed");
				} else {
					steps.setText(solution.getLength() + " steps");
				}
				while (solution != null) {
					cells.get(solution.getEnd()).setFill(Color.BLUE);
					solution = solution.getPrev();
				}
			} catch (InstantiationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	Searcher<Trail> makeSearcher() throws InstantiationException, IllegalAccessException {
		if (stackButton.isSelected()) {
			String className = stackBox.getSelectionModel().getSelectedItem();
			return new WrappedStack<Trail>(stackFinder.newInstanceOf(className));
		} else {
			String className = queueBox.getSelectionModel().getSelectedItem();
			return new WrappedQueue<Trail>(queueFinder.newInstanceOf(className));
		}
	}
	
	void updateCellCounts() {
		for (Cell type: Cell.values()) {
			cellCounts.get(type).setText(Integer.toString(mazeData.getCountOf(type)));
		}
	}
}
